// Copyright (c) 2023 Byron Lathi

module axi_crossbar #(
    parameter N_INITIATORS = 2,
    parameter N_TARGETS = 2,

    parameter AXI_ADDR_WIDTH = 32,
    parameter AXI_DATA_WIDTH = 32,
    parameter AXI_ID_W = 8

    // Add feature pipelining
    // add feature timeout
    // figure out how to parameterize this!
)(
    input clk,
    input rst,

    input   [0:N_INITIATORS-1]  [AXI_ADDR_WIDTH-1:0]    ini_araddr,
    input   [0:N_INITIATORS-1]                          ini_arvalid,
    output  [0:N_INITIATORS-1]                          ini_arready,
    output  [0:N_INITIATORS-1]  [AXI_DATA_WIDTH-1:0]    ini_rdata,
    output  [0:N_INITIATORS-1]  [1:0]                   ini_rresp,
    output  [0:N_INITIATORS-1]                          ini_rvalid,
    input   [0:N_INITIATORS-1]                          ini_rready,
    input   [0:N_INITIATORS-1]  [AXI_ADDR_WIDTH-1:0]    ini_awaddr,
    input   [0:N_INITIATORS-1]                          ini_awvalid,
    output  [0:N_INITIATORS-1]                          ini_awready,
    input   [0:N_INITIATORS-1]  [AXI_DATA_WIDTH-1:0]    ini_wdata,
    input   [0:N_INITIATORS-1]                          ini_wvalid,
    output  [0:N_INITIATORS-1]                          ini_wready,
    input   [0:N_INITIATORS-1]  [AXI_ADDR_WIDTH/8-1:0]  ini_wstrb,
    output  [0:N_INITIATORS-1]  [1:0]                   ini_bresp,
    output  [0:N_INITIATORS-1]                          ini_bvalid,
    input   [0:N_INITIATORS-1]                          ini_bready,

    output  [0:N_TARGETS-1]     [AXI_ADDR_WIDTH-1:0]    tgt_araddr,
    output  [0:N_TARGETS-1]                             tgt_arvalid,
    input   [0:N_TARGETS-1]                             tgt_arready,
    input   [0:N_TARGETS-1]     [AXI_DATA_WIDTH-1:0]    tgt_rdata,
    input   [0:N_TARGETS-1]     [1:0]                   tgt_rresp,
    input   [0:N_TARGETS-1]                             tgt_rvalid,
    output  [0:N_TARGETS-1]                             tgt_rready,
    output  [0:N_TARGETS-1]     [AXI_ADDR_WIDTH-1:0]    tgt_awaddr,
    output  [0:N_TARGETS-1]                             tgt_awvalid,
    input   [0:N_TARGETS-1]                             tgt_awready,
    output  [0:N_TARGETS-1]     [AXI_DATA_WIDTH-1:0]    tgt_wdata,
    output  [0:N_TARGETS-1]                             tgt_wvalid,
    input   [0:N_TARGETS-1]                             tgt_wready,
    output  [0:N_TARGETS-1]     [AXI_ADDR_WIDTH/8-1:0]  tgt_wstrb,
    input   [0:N_TARGETS-1]     [1:0]                   tgt_bresp,
    input   [0:N_TARGETS-1]                             tgt_bvalid,
    output  [0:N_TARGETS-1]                             tgt_bready
);

logic [N_INITIATORS-1:0] w_req;

logic [0:N_INITIATORS-1][AXI_ADDR_WIDTH-1:0]    ini_internal_addr;
logic [0:N_INITIATORS-1][AXI_DATA_WIDTH-1:0]    ini_internal_w_data;
logic [0:N_INITIATORS-1][AXI_DATA_WIDTH/8-1:0]  ini_internal_wstrb;
logic [AXI_DATA_WIDTH-1:0]                      ini_internal_r_data;
logic [0:N_INITIATORS-1]                        ini_internal_rwb;
logic [0:N_INITIATORS-1]                        ini_internal_addr_valid;
logic [0:N_INITIATORS-1]                        ini_internal_r_valid;
logic [0:N_INITIATORS-1]                        ini_internal_w_done;

generate 

for (genvar i = 0; i < N_INITIATORS; i++) begin : l_initiators

    axi_slave u_axi_slave (
        .i_clk(clk),
        .i_rst(rst),

        .i_AWVALID      (ini_awvalid[i]),
        .o_AWREADY      (ini_awready[i]),
        .i_AWADDR       (ini_awaddr[i]),
        .i_AWPROT       ('0),

        .i_WVALID       (ini_wvalid[i]),
        .o_WREADY       (ini_wready[i]),
        .i_WDATA        (ini_wdata[i]),
        .i_WSTRB        (ini_wstrb[i]),

        .o_BVALID       (ini_bvalid[i]),
        .i_BREADY       (ini_bready[i]),
        .o_BRESP        (ini_bresp[i]),

        .i_ARVALID      (ini_arvalid[i]),
        .o_ARREADY      (ini_arready[i]),
        .i_ARADDR       (ini_araddr[i]),
        .i_ARPROT       ('0),

        .o_RVALID       (ini_rvalid[i]),
        .i_RREADY       (ini_rready[i]),
        .o_RDATA        (ini_rdata[i]),
        .o_RRESP        (ini_rresp[i]),

        .o_addr         (ini_internal_addr[i]),
        .o_w_data       (ini_internal_w_data[i]),
        .o_wstrb        (ini_internal_wstrb[i]),
        .i_r_data       (ini_internal_r_data),
        .o_rwb          (ini_internal_rwb[i]),
        .o_addr_valid   (ini_internal_addr_valid[i]),
        .i_r_valid      (ini_internal_r_valid[i]),
        .i_w_done       (ini_internal_w_done[i])
    );

end

endgenerate

assign w_req = ini_internal_addr_valid;

logic [$clog2(N_INITIATORS)-1:0] master_sel;
logic master_valid;

rr_scheduler #(
    .N_INITIATORS(N_INITIATORS),
    .N_TARGETS(N_TARGETS)
) u_scheduler (
    .clk(clk),
    .rst(rst),

    .ini_req(w_req),
    .ini_en(),
    .master_sel(master_sel),
    .master_valid(master_valid)
);

logic [AXI_ADDR_WIDTH-1:0] selected_address;
logic [AXI_ADDR_WIDTH-1:0] decoded_address;
logic [$clog2(N_TARGETS)-1:0] target_sel;
logic target_valid;

logic [AXI_DATA_WIDTH-1:0] tgt_internal_w_data;
logic [AXI_DATA_WIDTH/8-1:0] tgt_internal_wstrb;
logic tgt_internal_rwb;

slave_addr_decoder #(
    .N_TARGETS(N_TARGETS)
) u_addr_decoder(
    .global_addr(selected_address),
    .local_addr(decoded_address),

    .master_valid(master_valid),
    .target_sel(target_sel),
    .target_valid(target_valid)
);

logic [N_TARGETS-1:0] tgt_sel;

logic [AXI_DATA_WIDTH-1:0] tgt_internal_rdata [N_TARGETS];
logic tgt_internal_r_valid [N_TARGETS];
logic tgt_internal_w_done [N_TARGETS];
logic [0:N_TARGETS-1] tgt_internal_addr_valid;


//TODO fix this
assign selected_address = ini_internal_addr[master_sel];
assign tgt_internal_rwb = ini_internal_rwb[master_sel];

always_comb begin

    tgt_internal_w_data = ini_internal_w_data[master_sel];
    tgt_internal_wstrb = ini_internal_wstrb[master_sel];

    if (target_valid) begin
        for (int i = 0; i < N_TARGETS; i++) begin
            if (i == target_sel) begin
                ini_internal_r_data = tgt_internal_rdata[i];
                ini_internal_r_valid[master_sel] = tgt_internal_r_valid[i];
                ini_internal_w_done[master_sel] = tgt_internal_w_done[i];
                tgt_internal_addr_valid[i] = ini_internal_addr_valid[master_sel];
            end else begin
                tgt_internal_addr_valid[i] = '0;
            end
        end
    end else begin
        ini_internal_r_data = '0;
    end
end

generate

for (genvar i = 0; i < N_TARGETS; i++) begin : l_targets

    axi_master u_axi_master (
        .i_clk(clk),
        .i_rst(rst),

        .o_AWVALID      (tgt_awvalid[i]),
        .i_AWREADY      (tgt_awready[i]),
        .o_AWADDR       (tgt_awaddr[i]),
        .o_AWPROT       (),
        .o_WVALID       (tgt_wvalid[i]),
        .i_WREADY       (tgt_wready[i]),
        .o_WDATA        (tgt_wdata[i]),
        .o_WSTRB        (tgt_wstrb[i]),
        .i_BVALID       (tgt_bvalid[i]),
        .o_BREADY       (tgt_bready[i]),
        .i_BRESP        (tgt_bresp[i]),
        .o_ARVALID      (tgt_arvalid[i]),
        .i_ARREADY      (tgt_arready[i]),
        .o_ARADDR       (tgt_araddr[i]),
        .o_ARPROT       (),
        .i_RVALID       (tgt_rvalid[i]),
        .o_RREADY       (tgt_rready[i]),
        .i_RDATA        (tgt_rdata[i]),
        .i_RRESP        (tgt_rresp[i]),

        .i_addr         (decoded_address),
        .i_w_data       (tgt_internal_w_data),
        .i_wstrb        (tgt_internal_wstrb),
        .o_r_data       (tgt_internal_rdata[i]),
        .i_rwb          (tgt_internal_rwb),
        .i_addr_valid   (tgt_internal_addr_valid[i]),
        .o_r_valid      (tgt_internal_r_valid[i]),
        .o_w_done       (tgt_internal_w_done[i])
    );

end

endgenerate

endmodule