module axi_master #(
    parameter AXI_ADDR_WIDTH = 32,
    parameter AXI_DATA_WIDTH = 32
)(
    input i_clk,
    input i_rst,


    /* AXI4-LITE master interface */
    output  logic                           o_AWVALID,
    input                                   i_AWREADY,
    output  logic   [AXI_ADDR_WIDTH-1:0]    o_AWADDR,
    output  logic   [2:0]                   o_AWPROT,

    output  logic                           o_WVALID,
    input                                   i_WREADY,
    output  logic   [AXI_DATA_WIDTH-1:0]    o_WDATA,
    output  logic   [AXI_DATA_WIDTH/8-1:0]  o_WSTRB,

    input                                   i_BVALID,
    output  logic                           o_BREADY,
    input           [1:0]                   i_BRESP,

    output  logic                           o_ARVALID,
    input                                   i_ARREADY,
    output  logic   [AXI_ADDR_WIDTH-1:0]    o_ARADDR,
    output  logic   [2:0]                   o_ARPROT,

    input                                   i_RVALID,
    output  logic                           o_RREADY,
    input           [AXI_DATA_WIDTH-1:0]    i_RDATA,
    input           [1:0]                   i_RRESP,

    /* Internal interface */
    input           [AXI_ADDR_WIDTH-1:0]    i_addr,
    input           [AXI_DATA_WIDTH-1:0]    i_w_data,
    input           [AXI_DATA_WIDTH/8-1:0]  i_wstrb,
    output  logic   [AXI_DATA_WIDTH-1:0]    o_r_data,
    input                                   i_rwb,
    input                                   i_addr_valid,
    output  logic                           o_r_valid,
    output  logic                           o_w_done
);

enum logic [1:0] {IDLE, READ_ADDR, READ_DATA, WRITE} state, state_next;

logic [AXI_DATA_WIDTH-1:0] r_w_data, r_w_data_next;
logic [AXI_ADDR_WIDTH-1:0] r_addr, r_addr_next;
logic [AXI_DATA_WIDTH/8-1:0] r_wstrb, r_wstrb_next;

always_comb begin
    state_next = state;
    o_r_valid = '0;

    o_WVALID = '0;
    o_AWVALID = '0;
    o_ARVALID = '0;
    o_AWADDR = '0;
    o_ARADDR = '0;
    o_WSTRB = '0;
    o_WDATA = '0;
    o_BREADY = '0;
    o_RREADY = '0;

    o_w_done = '0;
    o_r_data = '0;
    o_r_valid = '0;
    r_w_data_next = r_w_data;
    r_addr_next = r_addr;
    r_wstrb_next = r_wstrb;

    case (state)
        IDLE: begin
            if (i_addr_valid) begin
                if (i_rwb) begin
                    state_next = READ_ADDR;
                end else begin
                    r_w_data_next = i_w_data;
                    r_addr_next = i_addr;
                    r_wstrb_next = i_wstrb;
                    state_next = WRITE;
                end
            end
        end

        READ_ADDR: begin
            o_ARVALID = '1;

            if (o_ARVALID & i_ARREADY) begin
                state_next = READ_DATA;
            end

            o_ARADDR = i_addr;
        end

        READ_DATA: begin
            if (i_RVALID) begin
                state_next = IDLE;
                o_r_data = i_RDATA;
                o_r_valid = '1;
            end

            o_RREADY = '1;
        end

        WRITE: begin
            o_AWVALID = '1;
            o_AWADDR = r_addr;
            o_WVALID = '1;
            o_WSTRB = r_wstrb;
            o_WDATA = r_w_data;
            o_BREADY = '1;

            if (i_BVALID) begin
                state_next = IDLE;
                o_w_done = '1;
            end
        end

        default: begin

        end
    endcase
end

always @(posedge i_clk or posedge i_rst) begin
    if (i_rst) begin
        state <= IDLE;
        r_w_data <= '0;
        r_addr <= '0;
        r_wstrb <= '0;
    end else begin
        state <= state_next;
        r_w_data <= r_w_data_next;
        r_wstrb <= r_wstrb_next;
        r_addr <= r_addr_next;
    end
end

endmodule