/*
 *  axi_slave.sv
 *
 *  Combines the multiple axi channels into a single output
 * 
 */

module axi_slave #(
    parameter AXI_ADDR_WIDTH = 32,
    parameter AXI_DATA_WIDTH = 32
)(
    input i_clk,
    input i_rst,


    /* AXI4-LITE slave interface */
    input i_AWVALID,
    output logic o_AWREADY,
    input [AXI_ADDR_WIDTH-1:0] i_AWADDR,
    input [2:0] i_AWPROT,

    input i_WVALID,
    output logic o_WREADY,
    input [AXI_DATA_WIDTH-1:0] i_WDATA,
    input [AXI_DATA_WIDTH/8-1:0] i_WSTRB,

    output logic o_BVALID,
    input i_BREADY,
    output logic [1:0] o_BRESP,

    input i_ARVALID,
    output logic o_ARREADY,
    input [AXI_ADDR_WIDTH-1:0] i_ARADDR,
    input [2:0] i_ARPROT,

    output logic o_RVALID,
    input i_RREADY,
    output logic [AXI_DATA_WIDTH-1:0] o_RDATA,
    output logic [1:0] o_RRESP,

    /* Internal interface */
    output logic [AXI_ADDR_WIDTH-1:0] o_addr,
    output logic [AXI_DATA_WIDTH-1:0] o_w_data,
    output logic [AXI_DATA_WIDTH/8-1:0] o_wstrb,
    input [AXI_DATA_WIDTH-1:0] i_r_data,
    output logic o_rwb,
    output logic o_addr_valid,
    input i_r_valid,
    input i_w_done
);

logic [AXI_ADDR_WIDTH-1:0] addr, addr_next;
logic [AXI_DATA_WIDTH-1:0] w_data, w_data_next;
logic [AXI_DATA_WIDTH-1:0] r_data, r_data_next;
logic rwb;
logic addr_valid, addr_valid_next;
logic data_valid, data_valid_next;
logic [AXI_DATA_WIDTH/8-1:0] wstrb, wstrb_next;

enum logic [1:0] {IDLE, READ_WAIT, READ, WRITE} state, state_next;

always_comb begin
    state_next = state;

    o_addr = '0;
    o_w_data = '0;
    o_rwb = '1;
    o_addr_valid = '0;
    o_wstrb = '0;
    wstrb_next = wstrb;

    o_BRESP = '0;
    o_RVALID = '0;
    o_ARREADY = '0;
    o_RDATA = '0;
    o_RRESP = '0;
    o_WREADY = '0;

    addr_valid_next = '0;
    data_valid_next = '0;
    r_data_next = r_data;
    w_data_next = w_data;
    addr_next = addr;

    case (state)
        IDLE: begin
            o_RVALID = '0;
            o_BVALID = '0;
            if (i_ARVALID) begin
                addr_next = i_ARADDR;
                o_ARREADY = '1;
                state_next = READ_WAIT;
            end

            if (i_AWVALID) begin
                addr_next = i_AWADDR;
                addr_valid_next = '1;
            end

            if (i_WVALID) begin
                w_data_next = i_WDATA;
                wstrb_next = i_WSTRB;
                data_valid_next = '1;
            end

            if (addr_valid & data_valid) begin
                state_next = WRITE;
            end
        end

        READ_WAIT: begin
            o_BVALID = '0;
            o_addr = addr;
            o_addr_valid = '1;
            o_rwb = '1;
            if (i_r_valid) begin
                r_data_next = i_r_data;
                state_next = READ;
            end
        end

        READ: begin
            o_BVALID = '0;
            o_RVALID = '1;
            o_RDATA = r_data;
            o_RRESP = '0;
            // master wait for slaves?
            if (i_RREADY) begin
                state_next = IDLE;
            end
        end

        WRITE: begin
            o_addr = addr;
            o_addr_valid = '1;
            o_w_data = w_data;
            o_rwb = '0;
            o_wstrb = wstrb;
    
            o_BVALID = i_w_done;
            o_BRESP = '0;
            // master wait for slaves?
            if (i_BREADY & i_w_done) begin
                state_next = IDLE;
                addr_valid_next = '0;
                data_valid_next = '0;
            end
        end

        default: begin
            state_next = IDLE;
        end
    endcase
end

always @(posedge i_clk or posedge i_rst) begin
    if (i_rst) begin
        state <= IDLE;
        addr <= '0;
        w_data <= '0;
        data_valid <= '0;
        wstrb <= '0;
        r_data <= '0;
    end else begin
        addr <= addr_next;
        r_data <= r_data_next;
        state <= state_next;
        w_data <= w_data_next;
        wstrb <= wstrb_next;
        data_valid <= data_valid_next;
        addr_valid <= addr_valid_next;
    end
end

endmodule
